/***********************************************************************

xepr2gfac - convert xepr DTA/DSC data to comma separated values (.csv) with Field or g-factor x-axis

COMPILE:

gcc -o xepr2gfac xepr2gfac.c

USAGE:
	xepr2gfac xepr_source_file
	
***********************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <sys/types.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

/**********************************************************************/
/* I've included several other parameters that may be useful one day  */
/**********************************************************************/

typedef struct
{
int xpts;
int ypts;
double xmin;
double xwid;
double mwfq;
char bseq[8];
char titl[256];
char irnam[16];
char xnam[8];
char iruni[8];
char xuni[8];
} epr_parameters;

int IS_BIG_ENDIAN=0;

/**********************************************************************/
main(argc, argv)
int	argc;
char	**argv;
{

	FILE *f_data;
	FILE *f_param;
	FILE *f_csv;
	char basefilename[256], datafilename[256];
	char paramfilename[256], csvfilename[256];
        extern char *strerror();
        extern int  errno;
	int np, xpts=1024;
	epr_parameters *parameters;

/* check correct number of arguments */

	if (argc != 2) {
		(void) printf("... wrong number of arguments\n");
		(void) printf(
		"USAGE: %s xepr_source_file\n",
		argv[0]);
		(void) printf("EXIT\n\n");
		return(1);
	} 

/* copy filename from command line to basefilename */
	sprintf(basefilename,"%s",argv[1]);

/* remove any DAT or DSC suffix */
	remove_suffix(basefilename,".DSC");
	remove_suffix(basefilename,".DTA");

/* copy basefilename plus appropriate suffix to data, csv and param files */
	sprintf(datafilename,"%s%s",basefilename,".DTA");
	sprintf(paramfilename,"%s%s",basefilename,".DSC");
	sprintf(csvfilename,"%s%s",basefilename,".csv");

/* check data file can be opened */

        if ((f_data = fopen(datafilename, "rb")) == NULL) {
                (void) printf("... cannot read %s: \n", datafilename);
                (void) printf("    ERROR %s\n", strerror(errno));
                (void) printf("EXIT\n\n");
                return(1);
        }

/* check param file can be opened */

        if ((f_param = fopen(paramfilename, "r")) == NULL) {
                (void) printf("... cannot read %s: \n", paramfilename);
                (void) printf("    ERROR %s\n", strerror(errno));
                (void) printf("EXIT\n\n");
                return(1);
        }

/* check output file can be opened */

        if ((f_csv = fopen(csvfilename, "w")) == NULL) {
                (void) printf("... cannot write to %s: \n", csvfilename);
                (void) printf("    ERROR %s\n", strerror(errno));
                (void) printf("EXIT\n\n");
                return(1);
        }

/* Allocate memory for parameters */
	parameters = (epr_parameters *) malloc(sizeof(epr_parameters));
	if (parameters == NULL)
    		{
    		(void) fprintf(stderr,"%s: Error allocating memory\n",argv[0]);
    		return(1);
    		}

/* read parameters from DSC file */
	read_epr_parameters(parameters,f_param);

	np = convert(f_data, parameters, f_csv);

/* Check return value of convert, 0=failure, 1=OK, 2= multidimensional data */
	if (np==0)
		fprintf(stderr,"*** Conversion Failed ***\n");
	else 
		fprintf(stderr,"*** Conversion succeeded rows=%d ***\n",np);
	
/* close input/ouput files */

	(void)fclose(f_data);
	(void)fclose(f_param);
	(void)fclose(f_csv);

/* free allocated memory */

	free(parameters);

	return(0);
}


/**********************************************************************/
int
convert (f_data, parameters, f_csv)
FILE 	*f_data;
epr_parameters *parameters;
FILE	*f_csv;
/*----------------------------------------------------------------------
Convert DTA file to ascii format file.
----------------------------------------------------------------------*/
{
	void	reverse8();

	char	**lb;
	double 	*ip;
	size_t	cnt;
	int	i, m;
        extern char *strerror();
        extern int  errno;

/* allocate memory for xpts data words */
	ip = (double*) malloc((unsigned)(parameters->xpts * sizeof(double)));
	if (ip == NULL)
		{
		(void) printf("... no memory (%d bytes)\n", 
			parameters->xpts * sizeof(double));
		(void) printf("EXIT\n\n");
		return(0);
		}

/* allocate memory for xpts * ypts columns */
/* assume each column requires 16 bytes */

	lb = malloc(parameters->xpts * sizeof(*lb));	/* 1st dim malloc */
        if (lb == NULL)
		{
                (void) printf("... no memory (%d bytes)\n",
                        parameters->xpts * sizeof *lb);
                (void) printf("EXIT\n\n");
                return(0);
        	}
	else
		{
		for (m = 0 ; m < parameters->xpts ; m++) /* 2nd dim malloc */
			{
			/* we add 3 to parameters->ypts because ypts is 0  */
			/* for 1D data, also we need two extra columns for y */
			lb[m] = malloc((parameters->ypts + 3) * 24);
			if (lb[m] == NULL)
				{
				(void) printf("... no memory (%d bytes)\n",
                        		(parameters->ypts * 2) * 24);
                		(void) printf("EXIT\n\n");
                		return(0);
				}
			}
                }

/* convert the data */
	i = 0;
	do {
	     if ( i==0 )
		{
		cnt = fread((char*)ip,sizeof(double),parameters->xpts,f_data);
		if (cnt == parameters->xpts)
			{
			++i;
			for (m = 0; m < parameters->xpts; m++)
			    {
			    if (IS_BIG_ENDIAN)
				reverse8(&ip[m]);
			    double x_value=parameters->xmin+((parameters->xwid/(parameters->xpts-1))*m);
			    if (parameters->ypts == 0)
				{
				/* formula for g-value taken from the 
				* National High Magnetic Field Laboratory
				* Tallahassee */
			        double g_value= (714.4775 * parameters->mwfq) / (x_value * 1000000000);
			        sprintf(lb[m],"%0.9f,%0.9f,%0.9f", x_value, g_value, ip[m]);
				}
			    else
				sprintf(lb[m],"%0.9f,%0.9f", x_value, ip[m]);
  			    }
			}
		}
	     else
		{
		cnt = fread((char*)ip,sizeof(double),parameters->xpts,f_data);
		if (cnt == parameters->xpts)
			{
			++i;
			for (m = 0; m < parameters->xpts; m++)
			    {
			    if (IS_BIG_ENDIAN)
				reverse8(&ip[m]);
			    sprintf(lb[m],"%s,%0.5f",lb[m],ip[m]);
  			    }
			}
		}
	    } while (cnt == parameters->xpts);

/* check there's nothing left over */

	if (cnt != 0)
	  {
	  (void) printf(
	  "... %d bytes remaining, incomplete conversion (?).\n",cnt);
          (void) printf("FILE ERROR %s\n", strerror(errno));
	  }

	for (m = 0; m < parameters->xpts; m++)
    	    fprintf(f_csv,"%s\n",lb[m]);

/* free allocated memory */
	for (m = 0; m < parameters->xpts; m++)
	    free(lb[m]);
	free(lb);
	free(ip);
	return i;
}


/**********************************************************************/
void
reverse8(datapointer)

char    *datapointer;
{
  double *dp;
  char *ca;

  union {
    char    ch[8];
    double    d;
  }       bs;

  ca = (char *) datapointer;
  dp = (double *) datapointer;
    bs.ch[7] = *ca++;
    bs.ch[6] = *ca++;
    bs.ch[5] = *ca++;
    bs.ch[4] = *ca++;
    bs.ch[3] = *ca++;
    bs.ch[2] = *ca++;
    bs.ch[1] = *ca++;
    bs.ch[0] = *ca++;
    *dp++ = bs.d;

}

/**********************************************************************/
remove_suffix (char *name, const char *suffix)
{
  char *np;
  const char *sp;

  np = name + strlen (name);
  sp = suffix + strlen (suffix);

  while (np > name && sp > suffix)
    if (*--np != *--sp)
      return;
  if (np > name)
    *np = '\0';
}

/**********************************************************************/
read_epr_parameters(parameters, f_param)
epr_parameters *parameters;
FILE *f_param;
{
char str[200], *str1, *str2;

/* set ypts=0 in case YPTS is not found in file */
parameters->ypts=0;

while(fgets(str,sizeof(str),f_param) != NULL)
  {
  /* strip trailing '\n' if it exists */
  int len = strlen(str)-1;
  if(str[len] == '\n')
    str[len] = 0;
  str1 = strtok(str, " 	");
  if (str1 == NULL)
    break;

/* this is horrible, but a quick and easy way to parse the parameter file! */

  if (strcmp(str1,"BSEQ")==0) /* BSEQ is the byte order parameter */
    {
    /* extract next string from string sequence */
    str2 = strtok(NULL, "   ");
    sprintf(parameters->bseq,"%s",str2);
    if (strcmp(parameters->bseq,"BIG")==0)
        IS_BIG_ENDIAN=1;
    }
  else if (strcmp(str1,"TITL")==0)
    {
    /* extract next string from string sequence */
    while (1)
       {
       /* check if there is nothing else to extract */
       str2 = strtok(NULL, "       ");
       if (str2 == NULL)
          break;
       sprintf(parameters->titl,"%s %s",parameters->titl,str2);
       }
    }
  else if (strcmp(str1,"XNAM")==0)
    {
    /* extract next string from string sequence */
    str2 = strtok(NULL, "   ");
    sprintf(parameters->xnam,"%s",str2);
    }
  else if (strcmp(str1,"XUNI")==0)
    {
    /* extract next string from string sequence */
    str2 = strtok(NULL, "   ");
    sprintf(parameters->xuni,"%s",str2);
    }
  else if (strcmp(str1,"IRNAM")==0)
    {
    /* extract next string from string sequence */
    str2 = strtok(NULL, "   ");
    sprintf(parameters->irnam,"%s",str2);
    }
  else if (strcmp(str1,"IRUNI")==0)
    {
    /* extract next string from string sequence */
    str2 = strtok(NULL, "   ");
    sprintf(parameters->iruni,"%s",str2);
    }
  else if (strcmp(str1,"XPTS")==0)
    {
    /* extract next string from string sequence */
    str2 = strtok(NULL, "   ");
    parameters->xpts=atoi(str2);
    }
  else if (strcmp(str1,"YPTS")==0)
    {
    /* extract next string from string sequence */
    str2 = strtok(NULL, "   ");
    parameters->ypts=atoi(str2);
    }
  else if (strcmp(str1,"XMIN")==0)
    {
    /* extract next string from string sequence */
    str2 = strtok(NULL, "   ");
    parameters->xmin=atof(str2);
    }
  else if (strcmp(str1,"MWFQ")==0)
    {
    /* extract next string from string sequence */
    str2 = strtok(NULL, "   ");
    parameters->mwfq=atof(str2);
    }
  else if (strcmp(str1,"XWID")==0)
    {
    /* extract next string from string sequence */
    str2 = strtok(NULL, "   ");
    parameters->xwid=atof(str2);
    }
    /* loop until finishied */
    while (1)
       {
       /* check if there is nothing else to extract */
       str2 = strtok(NULL, "       ");
       if (str2 == NULL)
          break;
       }
  }
}
